package desmoj.tutorial2.StatisticsExample;

import desmoj.core.simulator.*;
import desmoj.core.dist.*;
import desmoj.core.statistic.*;
import desmoj.core.advancedModellingFeatures.*;
import java.util.concurrent.TimeUnit;

/**
 * This is the model class. It is the main class of a simple process-oriented
 * model of the railway station inside a container terminal using bins to represent
 * different storage areas.
 * Container ships arrive at the terminal and deliver containers to the storage
 * areas. From there they are loaded onto trains that leave for four different
 * destinations: Northern, Central, and Eastern Europe and Germany. Each
 * destination is allocated a storage area (Bin). Trains only load containers
 * from the storage area that is associated with their destination whereas
 * ships deliver containers to all storage areas.
 * In addition to the two counters from the CountExample model, several other
 * data collectors are added: a tally to record the average time a trains needs to
 * wait for containers, an accumulate to calculate the mean storage load level of the four
 * storage areas, a time series to log the storage load level of the Northern
 * Europe dedicated storage area, and a histogram to record a distribution
 * of ship sizes.
 * @author Olaf Neidhardt, Ruth Meyer
 */
public class StatisticsExample extends Model {

	/**
	 *	A random stream to determine the number of containers for each destination
	 */
	private DiscreteDistUniform containerDestinationStream;

	/**
	 *	A random stream to determine the time of arrival of the next ship
	 */
	private ContDistExponential shipArrivalStream;

	/**
	 *	A random stream to determine the length of a train
	 */
	private DiscreteDistUniform trainLengthStream;

	/** A Bin construct used to used to keep the Eastern Europe bound containers */
	protected Bin east;

	/** A Bin construct used to used to keep the Northern Europe bound containers */
	protected Bin north;

	/** A Bin construct used to used to keep the Central Europe bound containers */
	protected Bin central;

	/** A Bin construct used to used to keep the Germany bound containers */
	protected Bin germany;

	/** a counter to record the number of trains leaving the terminal */
	protected Count numberOfTrains;

	/** a counter to record the total number of containers exchanged */
	protected Count numberOfContainers;

	/** a tally to record the average time a ship needs to unload */
	protected Tally trainWaitTime;

	/** an accumulate to calculate the mean storage load level of all
	 *  four storage areas
	 */
	protected Accumulate storageLoadLevel;

	/** a time series to log the storage load level of the Northern
	 *  Europe dedicated storage area
	 */
	protected TimeSeries loadLevelNorth;

	/** a histogram to record the distribution of ship sizes */
	protected Histogram shipSize;

	/**
	 * This method constructs a new instance of a StatisticsExample Model
	 * Needed parameters are
	 * @param owner Model    the owning model, normally "null"
	 * @param name String   a name representing the model
	 * @param showInReport boolean    shall the model be represented in model report?
	 * @param showIntrace boolean     and in trace?
	 */
	public StatisticsExample(Model owner, String name, boolean showInReport, boolean showIntrace) {

		super(owner, name, showInReport, showIntrace);

	}
	/** returns a description of the model */
	public String description() {
		return "This is a simple process-oriented model of the railway station "+
		"inside a container terminal using bins to represent different storage areas. " +
		"Container ships arrive at the terminal and deliver containers to the storage " +
		"areas. From there they are loaded onto trains that leave for four different " +
		"destinations: Northern, Central, and Eastern Europe and Germany. Each " +
		"destination is allocated a storage area (Bin). Trains only load containers " +
		"from the storage area that is associated with their destination whereas " +
		"ships deliver containers to all storage areas. <br>" +
		"In addition to the two counters from the CountExample model, several other " +
		"data collectors are added: a tally to record the average time a trains needs to " +
		"wait for containers, an accumulate to calculate the mean storage load level of the four " +
		"storage areas, a time series to log the storage load level of the Northern " +
		"Europe dedicated storage area, and a histogram to record a distribution " +
		"of ship sizes.";
	}
	/**
	 * activates dynamic model components (simulation processes).
	 *
	 * This method is used to place all events or processes on the
	 * internal event list of the simulator which are necessary to start
	 * the simulation.
	 *
	 * In this case, the ship generator and the first four trains (one for each
	 * (destination) have to be created and activated.
	 */
	public void doInitialSchedules() {

		// create and activate the ship generator
		ShipGenerator generator = new ShipGenerator(this, "ShipGenerator", true);
		generator.activate(new TimeSpan(getShipArrivalTime(), TimeUnit.MINUTES));

		//create one train per direction
		//...and activate it too to ensure transportation
		Train easternBound = new Train(this, "Train to "+east.getName(), true, east);
		easternBound.activate();

		Train northernBound = new Train(this, "Train to "+north.getName(), true, north);
		northernBound.activate();

		Train centralBound = new Train(this, "Train to "+central.getName(), true, central);
		centralBound.activate();

		Train germanyBound = new Train(this, "Train to "+germany.getName(), true, germany);
		germanyBound.activate();

	}
	/**
	 * This method is used by the model components to receive
	 * a new sample out of a random stream.
	 * In this case it delivers a number of container containers.
	 * @return int
	 */
	protected long getNumberOfContainers() {
		return containerDestinationStream.sample();
	}
	/**
	 * This method is used by the model components to receive
	 * a new sample out of a random stream.
	 * In this case it delivers a new inter-arrival time for the
	 * arrival of the next ship.
	 * @return double
	 */
	protected double getShipArrivalTime() {
		return shipArrivalStream.sample();
	}
	/**
	 * This method is used by the model components to receive
	 * a new sample out of a random stream.
	 * In this case it delivers a train length (= number of container wagons)
	 * @return int
	 */
	protected long getTrainLength() {
		return trainLengthStream.sample();
	}

	/**
	 * This method is used by the model components to receive the current
	 * load level of all four storage areas combined, in order to be able
	 * to update the accumulate.
	 * @return long
	 */
	protected long getCurrentStorageLoad() {
		return (east.getAvail() + north.getAvail() + central.getAvail() + germany.getAvail());
	}

	/**
	 * initialises static model components like distributions and queues.
	 */
	public void init() {
		// initialise distributions
		// (1) for the number of containers a ship delivers to a destination
		containerDestinationStream = new DiscreteDistUniform(this,"No of containers per destination", 25,400, true, false);
		// (2) for the ships' interarrival time
		shipArrivalStream = new ContDistExponential(this,"Ships Arrival Stream", 6.0, true, false); ;
		// (3) for the length of trains (= number of containers they transport)
		trainLengthStream = new DiscreteDistUniform(this, "Train Length Stream", 30,90, true, false);

		// initialise the bins (storage areas), each with an initial number of 50 containers already stored in them
		east = new Bin(this, "Eastern Europe", 50, true, true);
		north = new Bin(this, "Northern Europe", 50, true, true);
		central = new Bin(this, "Central Europe", 50, true, true);
		germany = new Bin(this, "Germany", 50, true, true);

		// initialise the data collectors
		// (1) the two counters
		numberOfTrains = new Count(this, "number of trains", true, false);
		numberOfContainers = new Count(this, "number of containers", true, false);
		// (2) the tally
		trainWaitTime = new Tally(this, "train wait time", true, false);
		// (3) the accumulate
		storageLoadLevel = new Accumulate(this, "storage load level", true, false);
		// (4) the time series
		loadLevelNorth = new TimeSeries(this, "load level for Northern Europe", "StatisticsExample_LoadLevelNorth.txt", new TimeInstant(0), new TimeInstant(500), true, false);
		// (5) the histogram
		shipSize = new Histogram(this, "ship size", 200, 1400, 10, true, false);
	}

    /** run the model */
	public static void main(java.lang.String[] args) {

		// create model and experiment
		StatisticsExample model = new StatisticsExample(null, "StatisticsExample", true, false);
		Experiment exp = new Experiment("StatisticsExampleExperiment", TimeUnit.SECONDS, TimeUnit.MINUTES, null);

		// connect both
		model.connectToExperiment(exp);

		// set experiment parameters
		exp.setShowProgressBar(true);  // display a progress bar (or not)
		exp.stop(new TimeInstant(1500, TimeUnit.MINUTES));   // set end of simulation at 500 minutes
		exp.tracePeriod(new TimeInstant(0), new TimeInstant(100, TimeUnit.MINUTES));  // set the period of the trace
		exp.debugPeriod(new TimeInstant(0), new TimeInstant(50, TimeUnit.MINUTES));   // and debug output

		// start the Experiment with start time 0.0
		exp.start();

		// --> now the simulation is running until it reaches its ending criterion
		// ...
		// ...
		// <-- after reaching the ending criterion, the main thread returns here

		// generate the report (and other output files)
		exp.report();

		// stop all threads still alive and close all output files
		exp.finish();
	}
}